/**
  *  \file client/dialogs/fleetopts.cc
  *  \brief Fleet Cost Options dialog
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cctype>
#include "client/dialogs/fleetopts.h"
#include "gfx/palette.h"
#include "util/nls.h"
#include "client/wtile.h"
#include "gfx/gfxcontext.h"
#include "ui/uiglobal.h"
#include "cpluslib/misc.h"
#include "ui/widgets/button.h"
#include "ui/util.h"
#include "ui/layout.h"
#include "util/strtab.h"

namespace {
    class WFleetCostOptionsTile : public WTile {
     public:
        WFleetCostOptionsTile(WFleetCostOptions& opts);
        bool handleEvent(const UIEvent& event, bool second_pass);
        void getLayoutInfo(LayoutInfo& info);
        void drawData(GfxCanvas& can);

     private:
        void init();
        void computeLayout();

        WFleetCostOptions& opts;
        ObjectHolder h;
        int left_x, right_x, right_wi;
    };
}

static const char*const left_side[] = {
    _("Fighter cost"),
    _("Include torpedoes"),
    _("Include engines"),
    _("Include planet defense"),
    _("Include starbase"),
    _("Include tech levels"),
    _("Display by"),
};

static const char*const right_side[] = {
    _("no"),                    // 0 = false
    _("yes"),                   // 1 = true
    _("player"),                // 2 = 2+false
    _("team"),                  // 3 = 2+true
    _("not included"),          // 4 = 2+ff_Free
    _("built by \"lfm\""),      // 5 = 2+ff_LFM
    _("built on starbase"),     // 6 = 2+ff_Build
};

static const char keys[] = { 'F', 'T', 'E', 'P', 'B', 'L', 'Y' };

WFleetCostOptionsTile::WFleetCostOptionsTile(WFleetCostOptions& opts)
    : WTile(GfxPoint(1, 1)),
      opts(opts)
{
    init();
}

bool
WFleetCostOptionsTile::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case 'f':
            opts.fighters = WFleetCostOptions::Fighters(opts.fighters + 1);
            if (opts.fighters >= WFleetCostOptions::ff_MAX)
                opts.fighters = WFleetCostOptions::ff_Free;
            drawWidget();
            return true;
         case 't':
            opts.torps = !opts.torps;
            drawWidget();
            return true;
         case 'e':
            opts.engine = !opts.engine;
            drawWidget();
            return true;
         case 'p':
            opts.planet = !opts.planet;
            drawWidget();
            return true;
         case 'b':
            opts.base = !opts.base;
            drawWidget();
            return true;
         case 'l':
            opts.tech = !opts.tech;
            drawWidget();
            return true;
         case 'y':
            opts.team = !opts.team;
            drawWidget();
            return true;
        }
    }
    return WTile::handleEvent(event, second_pass);
}

void
WFleetCostOptionsTile::getLayoutInfo(LayoutInfo& info)
{
    info.min_size = info.pref_size = GfxPoint(right_x + 5, countof(left_side) * font_heights[FONT_TITLE]);
    info.flags = 0;
}

void
WFleetCostOptionsTile::drawData(GfxCanvas& can)
{
    const int h = font_heights[FONT_TITLE];
    const int y = h/2;

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);

    /* Left side */
    ctx.setTextAlign(0, 1);
    ctx.setColor(UIColor::tc_Static);
    for (uint_t i = 0; i < countof(left_side); ++i)
        outText(ctx, left_x, y + h*i, _(left_side[i]));

    /* Right side */
    ctx.setTextAlign(2, 1);
    ctx.setColor(UIColor::tc_Green);
    ctx.setSolidBackground();
    outTextF(ctx, right_x, y,     right_wi, _(right_side[4 + opts.fighters]));
    outTextF(ctx, right_x, y+h,   right_wi, _(right_side[opts.torps]));
    outTextF(ctx, right_x, y+2*h, right_wi, _(right_side[opts.engine]));
    outTextF(ctx, right_x, y+3*h, right_wi, _(right_side[opts.planet]));
    outTextF(ctx, right_x, y+4*h, right_wi, _(right_side[opts.base]));
    outTextF(ctx, right_x, y+5*h, right_wi, _(right_side[opts.tech]));
    outTextF(ctx, right_x, y+6*h, right_wi, _(right_side[2 + opts.team]));
}

void
WFleetCostOptionsTile::init()
{
    using std::tolower;
    computeLayout();
    setLayout(LeftJustified, 0, false);
    const int d = font_heights[FONT_TITLE];
    for (uint_t i = 0; i < countof(keys); ++i) {
        addControl(GfxRect(0, d*i+1, d-2, d-2), h.add(new UIButton(string_t(&keys[i], 1), tolower(keys[i]))));
    }
}

void
WFleetCostOptionsTile::computeLayout()
{
    int lw = getTextWidth(*fonts[FONT_NORMAL], StaticTranslatedStringTable(left_side, countof(left_side)));
    int rw = getTextWidth(*fonts[FONT_NORMAL], StaticTranslatedStringTable(right_side, countof(right_side)));

    left_x   = font_heights[FONT_TITLE] + 5;
    right_x  = left_x + lw + rw;
    right_wi = rw;
}

/** Edit fleet cost options.
    \param opts Option set to edit
    \retval true options were modified, user confirmed the dialog
    \retval false user canceled dialog, options not modified */
bool
editFleetCostOptions(WFleetCostOptions& opts)
{
    WFleetCostOptions new_opts(opts);

    ObjectHolder h;
    UIWindow& win = h.add(new UIWindow(_("Fleet Cost Comparison"), UIWindow::BLUE_WINDOW, COLOR_GRAY));
    win.setLayout(UIVBoxLayout::instance5);
    win.add(h.add(new WFleetCostOptionsTile(new_opts)));

    UIGroup& g = h.add(new UIGroup(UIHBoxLayout::instance5));
    g.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addStop());
    g.add(h.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE)).addStop());
    g.add(h.add(new UISpacer()));
    win.add(g);
    win.pack();
    ui_root->centerWidget(win);
    if (ui_root->runModal(win)) {
        opts = new_opts;
        return true;
    } else {
        return false;
    }
}
