/**
  *  \file client/globalact.h
  *  \brief Global Actions Core
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_CLIENT_GLOBALACT_H
#define PCC_V2_CLIENT_GLOBALACT_H

#include "game/objlist.h"
#include "cpluslib/object.h"
#include "cpluslib/sigslot.h"

/** Global action - execution part. A WGlobalActionExecuter lives
    during execution of a global action and determines the action to
    perform. */
class WGlobalActionExecuter : public Object {
 public:
    /** Execute for an object.
        \param obj Object. Can be anything (unowned, ufo, etc.), but
                   will always be an existing game object
        \param locked true to allow working on locked objects
        \retval true I worked on this object
        \retval false I ignored this object */
    virtual bool execute(GObject& obj, bool locked) = 0;
    /** Finish execution. Can display a success message.
        \param objs Number of objects handed to execute()
        \param success Number of objects acknowledged by execute() */
    virtual void finish(int objs, int success) = 0;
};

/** Global action - template part. An instance of a WGlobalAction
    descendant defines a global action. */
class WGlobalAction : public Object {
 public:
    /** Prepare global action. This can (and should) display a dialog
        to prepare an action, and then produce a WGlobalActionExecuter
        to execute it.
        \return newly-allocated WGlobalActionExecuter, or null to cancel */
    virtual WGlobalActionExecuter* prepare() = 0;
};

/** Global action parameters. Contains options for global option
    execution, and can use these for executing a WGlobalAction. */
class WGlobalActionParameter {
 public:
    enum Flag {
        ExceptNumericalFCodes = 1,
        ExceptSpecialFCodes   = 2,
        OnlySearchResult      = 4,
        OnlyMarked            = 8,
        OnlyShips             = 16,
        OnlyPlanets           = 32,
        OverrideLocks         = 64
    };

    WGlobalActionParameter(GObjectList& search_result);
    ~WGlobalActionParameter();

    bool hasFlag(int flag) const;
    void toggleFlag(int flag);
    void setFlag(int flag, bool value);
    GObjectList& getSearchResult();

    void execute(WGlobalAction& action);

    Signal0 sig_changed;

 private:
    void maybeProcess(int& objs, int& success, WGlobalActionExecuter& exec, GObject& obj);
    void maybeProcess(int& objs, int& success, WGlobalActionExecuter& exec, GObjectType& type);

    int flags;

    GObjectList& search_result;
};

#endif
