/**
  *  \file client/screens/shipscreen.cc
  *  \brief Ship Screen
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/screens/shipscreen.h"
#include "client/chart.h"
#include "client/cscreen.h"
#include "client/dialogs/baseammo.h"
#include "client/dialogs/clonedlg.h"
#include "client/dialogs/objselect.h"
#include "client/screens/fleetscreen.h"
#include "client/tiles.h"
#include "client/tiles/shipmission.h"
#include "client/tiles/shipoverview.h"
#include "client/widgets/fleetlist.h"
#include "client/widgets/objlist.h"
#include "client/wscanchart.h"
#include "cpluslib/format.h"
#include "game/config.h"
#include "game/formula.h"
#include "game/global.h"
#include "game/host.h"
#include "game/objsel.h"
#include "game/pref.h"
#include "game/ship.h"
#include "gfx/gfxcontext.h"
#include "gfx/palette.h"
#include "int/if/shipif.h"
#include "ui/msgbox.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "util/nls.h"

namespace {
    /// Ship scanner (with waypoint etc).
    class WShipScannerChartWidget : public WScannerChartWidget {
        GObjectSelection& selection;
        GShip* getShip();

     protected:
        void  lockQueryLocation(bool& hyperjumping, int& current_gravity_pid);
        int32 lockQueryDistance(GPoint pt);

     public:
        WShipScannerChartWidget(GObjectSelection& selection);
        ~WShipScannerChartWidget();
        void drawScanner(GfxCanvas& can);
        void drawPost(GfxCanvas& can);
        void userMoveScannerTo(GPoint pt);
    };

    /// Starship screen.
    class WShipScreen : public WControlScreen {
        WShipScannerChartWidget scanner_chart;

        void doShipBaseMenu(GShip& ship, UIBaseWidget& button);
        void doShipFleetMenu(GShip& ship, UIBaseWidget& button);
     public:
        WShipScreen(GObjectSelection& selection);
        ~WShipScreen();
        bool handleEvent(const UIEvent& e, bool second_pass);
        bool handleCommand(const UICommand& cmd, UIBaseWidget& sender);
        void update();
        virtual void enumContexts(ptr_vector<IntContext>& ctx);
    };
}

static ConfigIntOption opt_tie_waypoint(getUserPreferences(), "Scanner.TieWaypoint", ValueBoolParser::instance, 0);

int
chooseShip()
{
    if (current_ship_type.isEmpty()) {
        messageBox(_("You do not have any ships. "
                     "To build ships, use [B] on a starbase."),
                   _("Starships"));
        return 0;
    } else if (current_ship_type.isUnit()) {
        return current_ship_selection.getCurrentIndex();
    } else {
        GSimpleObjectSelection select;
        select.initFrom(current_ship_selection);
        WShipOverviewTile bot(select);
        WObjectSelectionDialog dlg(_("Select Ship"), "pcc2:unitsel", select);
        dlg.addWidget(bot);
        dlg.pack();
        ui_root->centerWidget(dlg);
        if (ui_root->runModal(dlg) == cm_Enter) {
            return select.getCurrentIndex();
        } else {
            return 0;
        }
    }
}

static bool
isPossibleFleetMember(const GShip& caller, const GShip& other)
{
    return other.getId() != caller.getId()
        && other.isPlayable(GObject::Playable)
        && other.getOwner() == caller.getOwner()
        && other.getFleetNumber() == 0
        && other.getPos() == caller.getPos();
}

/************************ WShipScannerChartWidget ************************/

WShipScannerChartWidget::WShipScannerChartWidget(GObjectSelection& selection)
    : selection(selection)
{ }

WShipScannerChartWidget::~WShipScannerChartWidget()
{ }

GShip*
WShipScannerChartWidget::getShip()
{
    return dynamic_cast<GShip*>(selection.getCurrentObject());
}

void
WShipScannerChartWidget::lockQueryLocation(bool& hyperjumping, int& current_gravity_pid)
{
    if (GShip* sh = getShip()) {
        hyperjumping        = sh->isHyperdriving();
        current_gravity_pid = getDisplayedTurn().getCurrentUniverse().getGravityPlanetAt(sh->getPos());
    } else {
        hyperjumping = false;
        current_gravity_pid = 0;
    }
}

int32
WShipScannerChartWidget::lockQueryDistance(GPoint pt)
{
    GShip* sh = getShip();
    if (sh != 0) {
        const int32_t dist2 = distanceSquared(sh->getPos(), pt);
        if (sh->isHyperdriving() && !host.isExactHyperjumpDistance2(dist2)) {
            return -1;
        } else {
            return dist2;
        }
    } else {
        return -1;
    }
}

void
WShipScannerChartWidget::drawScanner(GfxCanvas& can)
{
    if (GShip* sh = getShip()) {
        GChartViewport& vp = getViewport();
        GPoint wp = sh->getWaypoint();

        if (wp != getScannerTarget() && wp != getScannerOrigin()) {
            GfxPoint wp_screen = vp.scaleXY(wp);
            GfxContext ctx;
            ctx.useColorScheme(GfxStandardColorScheme::instance);
            ctx.useCanvas(can);
            ctx.setColor(COLOR_RED);
            drawLine(ctx, vp.scaleXY(getScannerOrigin()), wp_screen);
            vp.drawScannerEnd(ctx, wp_screen);
        }
    }

    WScannerChartWidget::drawScanner(can);
}

void
WShipScannerChartWidget::drawPost(GfxCanvas& can)
{
    if (GShip* sh = getShip()) {
        // Drawing parameters
        GfxContext ctx;
        const GChartViewport& vp = getViewport();

        ctx.useCanvas(can);
        ctx.useColorScheme(GfxStandardColorScheme::instance);
        ctx.useFont(*fonts[FONT_SMALL]);
        ctx.setTextAlign(1, 0);

        // Chunnel.
        // FIXME: PCC 1.x also parses when this ship goes through another ship's chunnel
        GChunnelData cd;
        if (parseChunnelMission(*sh, getDisplayedTurn().getCurrentUniverse(), cd)) {
            if (cd.failure == 0) {
                // no error
                ctx.setColor(COLOR_BLUEGRAY);
            } else if (cd.failure & ~cd.chf_MateAny) {
                // error on our side
                ctx.setColor(COLOR_RED);
            } else {
                // error on mate side
                ctx.setColor(COLOR_YELLOW);
            }
            ctx.setLinePattern(0xF0);
            drawLine(ctx,
                     vp.scaleXY(sh->getPos()),
                     vp.scaleXY(getDisplayedTurn().getCurrentUniverse().getShip(cd.target).getPos()));
            ctx.setLinePattern(SOLID_LINE);
        }

        // Draw mine laying info
        GMineMissionData md;
        if (parseMineLayingMission(*sh, getDisplayedTurn().getCurrentUniverse(), md)) {
            // We're laying a minefield. What kind?
            GMinefield* mf = 0;
            if (md.mine_id > 0)
                mf = getDisplayedTurn().getCurrentUniverse().ty_minefields.get(md.mine_id);

            // Figure out where it goes
            GPoint center;
            int radius;
            string_t notice;
            if (mf == 0) {
                // New minefield
                center = sh->getPos();
                radius = GMinefield::getRadiusFromUnits(md.num_units);
                notice = format(_("%d ly"), radius);
            } else {
                // Enlarge existing minefield
                center = mf->getPos();
                radius = GMinefield::getRadiusFromUnits(md.num_units + mf->getUnitsForLaying());
                notice = format(_("%d ly (+%d)"), radius, radius - mf->getRadius());
            }

            GfxPoint p = vp.scaleXY(center.getSimpleNearestAlias(vp.center));
            ctx.setColor(COLOR_YELLOW);

            // Limit check
            int limit = (md.type
                         ? config.MaximumWebMinefieldRadius(md.owner)
                         : config.MaximumMinefieldRadius(md.owner));
            if (radius > limit) {
                drawCircle(ctx, p, vp.scale(limit));
                ctx.setColor(COLOR_DARK);
                outText(ctx, p.x, p.y + 10 + font_heights[FONT_SMALL], _("<over limit>"));
            }

            // Draw circle
            drawCircle(ctx, p, vp.scale(radius));
            outText(ctx, p.x, p.y + 10, notice);
        } else if (parseMineScoopMission(*sh, getDisplayedTurn().getCurrentUniverse(), md)) {
            // We're scooping
            const GMinefieldType& mfc = getDisplayedTurn().getCurrentUniverse().ty_minefields;
            int room = sh->getFreeCargo();
            for (GMinefieldType::const_iterator mf = mfc.begin(); mf != mfc.end(); ++mf) {
                // Check whether we can scoop this field
                if (mf->getOwner() == md.owner
                    && distanceSquared(mf->getPos(), sh->getPos()) <= mf->getUnitsForLaying()
                    && (md.mine_id == 0 || md.mine_id == mf->getId()))
                {
                    // Okay, scoop it. First, figure out conversion rate.
                    int rate = ((mf->isWeb() ? config.UnitsPerWebRate(md.owner) : config.UnitsPerTorpRate(md.owner))
                                * square(sh->getTorpType()))
                        / 100;
                    if (rate == 0)
                        rate = 1;

                    // How much do we clear?
                    int32 units = mf->getUnitsForLaying();
                    int32 scoopedUnits;
                    int scoopedTorps = units / rate;
                    if (scoopedTorps > room) {
                        // We cannot clear this field
                        scoopedTorps = room;
                        scoopedUnits = room * rate;
                    } else {
                        // We can clear this field
                        scoopedUnits = units;
                    }

                    // Check limit
                    if (md.num_torps > 0 && md.num_torps < scoopedTorps) {
                        scoopedTorps = md.num_torps;
                        scoopedUnits = md.num_torps * rate;
                    }

                    // Draw it
                    if (scoopedTorps > 0) {
                        GfxPoint p = vp.scaleXY(mf->getPos().getSimpleNearestAlias(vp.center));
                        ctx.setColor(COLOR_YELLOW);

                        room -= scoopedTorps;
                        if (units > scoopedUnits) {
                            // Field stays
                            int radius = mf->getRadiusFromUnits(units - scoopedUnits);
                            drawCircle(ctx, p, radius);
                            outText(ctx, p.x, p.y + 10, format(_("%d ly (%d torp%!1{s%})"), radius, scoopedTorps));
                        } else {
                            // Field is gone
                            outText(ctx, p.x, p.y + 10, format(_("gone (%d torp%!1{s%})"), scoopedTorps));
                        }
                    }
                }
            }
        }
    }

    // Base class
    WScannerChartWidget::drawPost(can);
}

void
WShipScannerChartWidget::userMoveScannerTo(GPoint pt)
{
    setScannerTarget(pt);
    if (opt_tie_waypoint()) {
        sig_dblclick.raise(getId());
    }
}

/****************************** WShipScreen ******************************/

WShipScreen::WShipScreen(GObjectSelection& selection)
    : WControlScreen(selection, WScreenId::ShipScreen),
      scanner_chart(selection)
{
    createTileLayout(*this, "SHIPSCREEN", selection);
    addChart(scanner_chart);
    setKeymap(IntKeymap::getKeymapByName("SHIPSCREEN"));
}

WShipScreen::~WShipScreen()
{ }

void
WShipScreen::update()
{
    // FIXME: can be simplified using dynamic_cast
    int sid = getObjectSelection().getCurrentIndex();
    const GShip& ship = getObjectSelection().getCurrentUniverse()->getShip(sid);
    scanner_chart.setCenterPosition(ship.getPos());
    scanner_chart.setScannerOrigin(ship.getPos());
    scanner_chart.setScannerTarget(scanner_chart.clipScannerPosition(ship.getWaypoint()));
    scanner_chart.drawWidget();
}

void
WShipScreen::enumContexts(ptr_vector<IntContext>& ctx)
{
    ctx.push_back(new IntShipContext(getObjectSelection().getCurrentIndex()));
}

void
WShipScreen::doShipBaseMenu(GShip& ship, UIBaseWidget& button)
{
    enum {
        CancelShipyard,
        FixThisShip,
        RecycleThisShip,
        BuildAmmo,
        GoToBase,
        CloneShip
    };

    // Are we orbiting a base?
    // FIXME: this way of obtaining the universe and base is pretty indirect
    if (!current_base_type.isValidIndex(ship.getOrbitPlanetId())) {
        messageBox(_("We are not at one of our starbases."), _("Starbase Commands"));
        return;
    }
    GPlanet& base = dynamic_cast<GPlanet&>(current_base_type.getObjectByIndex(ship.getOrbitPlanetId()));
    GUniverse& univ = *current_base_type.getUniverseByIndex(ship.getOrbitPlanetId());

    // Build command menu
    UIStringList cmdlist;

    // - Fix/Recycle
    // FIXME: the "can we fix" logic is also in act-fix.cc; can we share it?
    int currentAction = base.getBaseShipyardId() == ship.getId() ? base.getBaseShipyardOrder() : 0;
    if (currentAction == sya_Fix) {
        cmdlist.add(CancelShipyard, _("Cancel \"repair\" order"));
    } else {
        cmdlist.add(FixThisShip, _("Fix (repair) this ship"));
    }
    if (currentAction == sya_Recycle) {
        cmdlist.add(CancelShipyard, _("Cancel \"recycle\" order"));
    } else {
        if (base.getOwner() == ship.getOwner())
            cmdlist.add(RecycleThisShip, _("Recycle this ship"));
    }

    // - Build torps/fighters
    if (ship.getNumBays() > 0) {
        cmdlist.add(BuildAmmo, _("Build fighters"));
    } else if (ship.getNumTorpLaunchers() > 0) {
        cmdlist.add(BuildAmmo, _("Build torpedoes"));
    }

    // - Starbase screen
    cmdlist.add(GoToBase, _("Starbase screen"));

    // - Clone
    cmdlist.add(CloneShip, _("Clone this ship"));

    // Show menu
    UIStandardListbox lb(0, 1, getTextWidth(*fonts[FONT_NORMAL], cmdlist) + 20, cmdlist);
    if (!lb.doStandardMenu(button))
        return;

    switch (int cmd = lb.getKey()) {
     case CancelShipyard:
        base.setBaseShipyardOrder(mp16_t(0), mp16_t(0));
        ship.markDirty();  // FIXME Q&D
        univ.doScreenUpdates();
        break;

     case FixThisShip:
     case RecycleThisShip:
        {
            bool accept = true;
            if (base.getBaseShipyardOrder() != sya_None && base.getBaseShipyardId() != ship.getId()) {
                // Find name of ship being repaired/recycled
                string_t shipName;
                if (univ.ty_any_ships.isValidIndex(base.getBaseShipyardId()))
                    shipName = univ.getShip(base.getBaseShipyardId()).getName(GObject::PlainName);
                else
                    shipName = format(_("Ship %d"), base.getBaseShipyardId());
                // Ask user
                if (base.getBaseShipyardOrder() == sya_Fix) {
                    accept = UIMessageBox(format(_("This starbase is already repairing %s. "
                                                   "If you continue, this order will be replaced. Continue?"),
                                                 shipName),
                                          _("Starbase Commands")).doYesNoDialog();
                } else if (base.getBaseShipyardOrder() == sya_Recycle) {
                    accept = UIMessageBox(format(_("This starbase is already recycling %s. "
                                                   "If you continue, this order will be replaced. Continue?"),
                                                 shipName),
                                          _("Starbase Commands")).doYesNoDialog();
                }
            }
            if (accept) {
                base.setBaseShipyardOrder(mp16_t(cmd == FixThisShip ? sya_Fix : sya_Recycle),
                                          mp16_t(ship.getId()));
                ship.markDirty();  // FIXME Q&D
                univ.doScreenUpdates();
            }
        }
        break;

     case BuildAmmo:
        doBaseBuildAmmo(base, ship);
        univ.doScreenUpdates();
        break;

     case GoToBase:
        postGoToScreen(*this, 3, ship.getOrbitPlanetId());
        break;

     case CloneShip:
        doCloneShip(ship, base);
        univ.doScreenUpdates();
        break;
    }
}

void
WShipScreen::doShipFleetMenu(GShip& ship, UIBaseWidget& button)
{
    enum {
        StartNewFleet,
        JoinFleet,
        LeaveFleet,
        GoToLeader,
        AddAll,
        FleetScreen
    };

    // Standard commands
    StringList menu;
    menu.add(FleetScreen, _("Fleet screen"));
    if (ship.getFleetNumber() == 0) {
        menu.add(StartNewFleet, _("Start new fleet"));
        if (!current_fleet_type.isEmpty()) {
            menu.add(JoinFleet, _("Join a fleet"));
        }
    } else {
        menu.add(LeaveFleet, _("Leave fleet"));
        if (!ship.isFleetLeader()) {
            menu.add(GoToLeader, _("Go to fleet leader"));
        }
    }

    // Only when there are matching ships
    GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
    for (int i = 1; i <= NUM_SHIPS; ++i) {
        if (isPossibleFleetMember(ship, univ.getShip(i))) {
            menu.add(AddAll, _("Add all to fleet"));
            break;
        }
    }

    // Menu
    UIStandardListbox lb(0, 1, getTextWidth(*fonts[FONT_NORMAL], menu) + 20, menu);
    if (!lb.doStandardMenu(button))
        return;

    int fid;
    switch (lb.getKey()) {
     case StartNewFleet:
        setFleetNumber(univ, ship.getId(), ship.getId());
        current_fleet_selection.setCurrentIndex(ship.getId());
        break;

     case JoinFleet:
        fid = doFleetList(_("Join Fleet"), univ, ship.getPos(), WFleetList::fAllFleets, 0);
        if (fid != 0) {
            setFleetNumber(univ, ship.getId(), fid);
        }
        break;

     case LeaveFleet:
        leaveFleet(univ, ship.getId());
        break;

     case GoToLeader:
        fid = ship.getFleetNumber();
        if (fid != 0) {
            postGoToScreen(*this, 1, fid);
        }
        break;

     case AddAll:
        fid = ship.getFleetNumber();
        if (fid == 0) {
            fid = ship.getId();
            setFleetNumber(univ, ship.getId(), fid);
        }
        for (int i = 1; i <= NUM_SHIPS; ++i) {
            const GShip& osh = univ.getShip(i);
            if (isPossibleFleetMember(ship, osh)) {
                setFleetNumber(univ, i, fid);
            }
        }
        current_fleet_selection.setCurrentIndex(fid);
        break;

     case FleetScreen:
        fid = ship.getFleetNumber();
        if (fid == 0) {
            fid = chooseFleet();
        }
        if (fid != 0) {
            postGoToScreen(*this, 10, fid);
        }
        break;
    }

    getDisplayedTurn().getCurrentUniverse().doScreenUpdates();
}

bool
WShipScreen::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        GObjectSelection& sel = getObjectSelection();
        switch (e.key.code) {
         case SDLK_TAB:
         case ss_Shift + SDLK_TAB:
         case ss_Ctrl + SDLK_TAB:
         case ss_Shift + ss_Ctrl + SDLK_TAB:
            if (GShip* ship = dynamic_cast<GShip*>(sel.getCurrentObject())) {
                GObjectList l;
                l.addObjectsAt(*sel.getCurrentUniverse(), ship->getPos(), 0, 0);
                l.postprocessUser(false);
                int cur = l.getIndexFor(*ship);
                int n = ((e.key.code & ss_Shift)
                         ? l.findPreviousIndexWrap(cur, (e.key.code & ss_Ctrl) != 0)
                         : l.findNextIndexWrap    (cur, (e.key.code & ss_Ctrl) != 0));
                if (n > 0) {
                    sel.setCurrentIndex(l.getObjectByIndex(n).getId());
                }
            }
            return true;

            /* ship screen keys:

               c-n              // list next turn
               q                // toggle independantmove
               m-r              // reset x/y */
        }
    }
    return WControlScreen::handleEvent(e, second_pass);
}

bool
WShipScreen::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    if (cmd.code == cm_Ship_BaseMenu) {
        if (GShip* sh = dynamic_cast<GShip*>(getObjectSelection().getCurrentObject())) {
            doShipBaseMenu(*sh, sender);
        }
        return true;
    } else if (cmd.code == cm_Ship_FleetMenu) {
        if (GShip* sh = dynamic_cast<GShip*>(getObjectSelection().getCurrentObject())) {
            doShipFleetMenu(*sh, sender);
        }
        return true;
    } else {
        return WControlScreen::handleCommand(cmd, sender);
    }
}

void
doShipScreen()
{
    WShipScreen(current_ship_selection).runScreen();
}
