/**
  *  \file client/widgets/scanresult.cc
  *  \brief WScanResult widget
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>
#include "client/widgets/scanresult.h"
#include "client/wscanchart.h"
#include "ui/table.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "game/planet.h"
#include "game/ship.h"
#include "client/tiles/csheader.h"
#include "game/global.h"
#include "ui/uiglobal.h"

/** Constructor. */
WScanResult::WScanResult(bool with_buttons)
    : chart(0),
      sid(0),
      xpid(0),
      gpid(0),
      with_buttons(with_buttons),
      button_l("L", 'l'),
      button_y("Y", 'y')
{
    setState(st_Disabled, true); /* we don't want focus */
    if (with_buttons) {
        addChildWidget(button_l, 0);
        addChildWidget(button_y, 0);
    }
}

WScanResult::~WScanResult()
{ }

/** Associate this scanner with a starchart widget.
    \param chart the starchart widget
    \pre this routine must not already have been called */
void
WScanResult::setChart(WScannerChartWidget& chart)
{
    ASSERT(!this->chart);
    this->chart = &chart;
    update_conn = chart.sig_moved.add(this, &WScanResult::update);
    update();
}

/** Update. This recomputes all displayed information, and requests a
    redraw of the widget if needed. */
void
WScanResult::update()
{
    if (!chart || !chart->isValid()) {
        sid = xpid = gpid = 0;
        if (UIBaseWidget* w = getParent())
            w->setStateOf(cm_CS_Xchg, st_Disabled, true);
    } else {
        const GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
        sid  = univ.getAnyShipAt(chart->getScannerTarget());
        xpid = univ.getPlanetAt(chart->getScannerTarget());
        if (xpid)
            gpid = xpid;
        else
            gpid = univ.getGravityPlanetAt(chart->getScannerTarget());
        if (UIBaseWidget* w = getParent())
            w->setStateOf(cm_CS_Xchg, st_Disabled, (xpid == 0 && gpid == 0 && sid == 0));
    }
    drawWidget(false);
}

void
WScanResult::drawContent(GfxCanvas& can)
{
    const GfxRect& e = getExtent();
    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());
    UITableWriter w(ctx, GfxRect(e.x, e.y, with_buttons ? e.w-47 : e.w, e.h), FONT_NORMAL);
    string_t scan, dist;

    w.newLColumn();
    w.writeFixedCell(UIColor::tc_Static, _("Scan:"));
    w.writeFixedCell(UIColor::tc_Static, _("Distance:"));

    w.newLColumn();
    if (chart) {
        if (!chart->isValid()) {
            scan = dist = "...";
        } else {
            // FIXME: can we share this with GGameTurn::getLocationName()?
            const GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
            if (xpid)
                scan = univ.getPlanet(xpid).getName(GObject::LongName);
            else if (sid)
                scan = univ.getShip(sid).getName(GObject::LongName);
            else if (gpid)
                scan = format(_("near %s"), univ.getPlanet(gpid).getName(GObject::LongName));
            else
                scan = format(_("Deep Space (%d,%d)"),
                              chart->getScannerTarget().x,
                              chart->getScannerTarget().y);
            dist = format(_("%.2f ly"),
                          std::sqrt(double(distanceSquaredRaw(chart->getScannerOrigin(),
                                                              chart->getScannerTarget()))));
        }
    }
    w.writeVariableCell(UIColor::tc_Green, scan);
    w.writeVariableCell(UIColor::tc_Green, dist);
    w.finish();
    can.update(getExtent());
}

void
WScanResult::getLayoutInfo(LayoutInfo& info)
{
    // FIXME?
    info.flags = 0;
    info.min_size = GfxPoint(240, 2*font_heights[FONT_NORMAL]);
    info.pref_size = GfxPoint(310, 2*font_heights[FONT_NORMAL]);
}

void
WScanResult::onResize(GfxRect /*oldSize*/, bool /*byLayout*/)
{
    if (with_buttons) {
        const GfxRect& e = getExtent();
        button_y.setExtent(GfxRect(e.getRightX()-22, e.y+2, 20, 20), true);
        button_l.setExtent(GfxRect(e.getRightX()-47, e.y+2, 20, 20), true);
    }
}
