/**
  *  \file reshack/tools.cc
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>
#include "reshack/tools.h"
#include "gfx/gfxcompl.h"
#include "gfx/gfxcontext.h"
#include "cpluslib/smartptr.h"
#include "gfx/pixmap.h"
#include "ui/ui.h"
#include "reshack/painter.h"

static Ptr<GfxPixmap> clipboard;

static inline int32_t square(int32_t x) { return x*x; }

static GfxRect
makeRectangle(GfxPoint a, GfxPoint b)
{
    GfxRect r;
    r.x = std::min(a.x, b.x);
    r.y = std::min(a.y, b.y);
    r.w = std::max(a.x, b.x) - r.x + 1;
    r.h = std::max(a.y, b.y) - r.y + 1;
    return r;
}


/********************************* RHTool ********************************/

RHTool::RHTool(bool needs_preview, string_t name)
    : needs_preview(needs_preview),
      name(name)
{ }

RHTool::~RHTool()
{ }

bool
RHTool::isUsable()
{
    return true;
}

/****************************** RHPencilTool *****************************/

void
RHPencilTool::click(GfxContext& c, GfxPoint pt, uint32_t /*bg*/)
{
    prev = pt;
    drawPixel(c, pt);
}

void
RHPencilTool::drag(GfxContext& c, GfxPoint pt)
{
    drawLine(c, prev, pt);
    prev = pt;
}

void
RHPencilTool::release(GfxContext& c, GfxPoint pt)
{
    drag(c, pt);
}

/****************************** RHBrushTool ******************************/

void
RHBrushTool::click(GfxContext& c, GfxPoint pt, uint32_t /*bg*/)
{
    drawFilledCircle(c, pt, 5);
}

void
RHBrushTool::drag(GfxContext& c, GfxPoint pt)
{
    drawFilledCircle(c, pt, 5);
}

void
RHBrushTool::release(GfxContext& c, GfxPoint pt)
{
    drawFilledCircle(c, pt, 5);
}

/******************************* RHLineTool ******************************/

void
RHLineTool::click(GfxContext& c, GfxPoint pt, uint32_t /*bg*/)
{
    prev = pt;
    drawPixel(c, pt);
}

void
RHLineTool::drag(GfxContext& c, GfxPoint pt)
{
    drawLine(c, prev, pt);
}

void
RHLineTool::release(GfxContext& c, GfxPoint pt)
{
    drag(c, pt);
}

/**************************** RHRectangleTool ****************************/

void
RHRectangleTool::click(GfxContext& c, GfxPoint pt, uint32_t /*bg*/)
{
    prev = pt;
    drawPixel(c, pt);
}

void
RHRectangleTool::drag(GfxContext& c, GfxPoint pt)
{
    drawRectangle(c, makeRectangle(pt, prev));
}

void
RHRectangleTool::release(GfxContext& c, GfxPoint pt)
{
    drag(c, pt);
}

/******************************* RHBarTool *******************************/

void
RHBarTool::click(GfxContext& c, GfxPoint pt, uint32_t /*bg*/)
{
    prev = pt;
    drawPixel(c, pt);
}

void
RHBarTool::drag(GfxContext& c, GfxPoint pt)
{
    c.setFillPattern(SOLID_FILL);
    drawBar(c, makeRectangle(pt, prev));
}

void
RHBarTool::release(GfxContext& c, GfxPoint pt)
{
    drag(c, pt);
}

/****************************** RHCircleTool *****************************/

void
RHCircleTool::click(GfxContext& c, GfxPoint pt, uint32_t /*bg*/)
{
    prev = pt;
    drawPixel(c, pt);
}

void
RHCircleTool::drag(GfxContext& c, GfxPoint pt)
{
    int radius = int(std::sqrt(double(square(pt.x - prev.x)) + double(square(pt.y - prev.y))) + 0.5);
    drawCircle(c, prev, radius);
}

void
RHCircleTool::release(GfxContext& c, GfxPoint pt)
{
    drag(c, pt);
}

/******************************* RHCopyTool ******************************/

void
RHCopyTool::click(GfxContext& c, GfxPoint pt, uint32_t bg)
{
    prev = pt;
    drawPixel(c, pt);
    bg_color = bg;
}

void
RHCopyTool::drag(GfxContext& c, GfxPoint pt)
{
    drawRectangle(c, makeRectangle(pt, prev));
}

void
RHCopyTool::release(GfxContext& c, GfxPoint pt)
{
    GfxPixmap* pix = dynamic_cast<GfxPixmap*>(&c.getCanvas());
    if (pix) {
        GfxRect area = makeRectangle(pt, prev);
        clipboard = new GfxPixmap(area.w, area.h, *pix);
        clipboard->setColorKey(bg_color);
        pix->blit(*clipboard, 0, 0, area);
        if (cut_notify)
            cut_notify->drawWidget();
    }
}

/****************************** RHPasteTool ******************************/

void
RHPasteTool::click(GfxContext& c, GfxPoint pt, uint32_t /*bg*/)
{
    // safety check
    if (clipboard != 0) {
        blitIndexed(c.getCanvas(), *clipboard, pt.x, pt.y);
    }
}

void
RHPasteTool::drag(GfxContext& c, GfxPoint pt)
{
    click(c, pt, 0);
}

void
RHPasteTool::release(GfxContext& c, GfxPoint pt)
{
    click(c, pt, 0);
}

bool
RHPasteTool::isUsable()
{
    return clipboard != 0;
}

/***************************** RHPipetteTool *****************************/

void
RHPipetteTool::click(GfxContext& c, GfxPoint pt, uint32_t /*bg*/)
{
    c.setRawColor(c.getCanvas().getPixel(pt));
}

void
RHPipetteTool::drag(GfxContext& c, GfxPoint pt)
{
    click(c, pt, 0);
}

void
RHPipetteTool::release(GfxContext& c, GfxPoint pt)
{
    click(c, pt, 0);
}

Ptr<GfxPixmap>
getClipboardImage()
{
    return clipboard;
}
